@@ -13,7 +13,8 @@
 
 \*---------------------------------------------------------------------------*/
 
-#include <iostream.h>
+#include <iostream>
+using namespace std;
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
